from abc import ABC, abstractmethod
from collections.abc import Generator, Iterable, MutableMapping, MutableSequence
from dataclasses import dataclass, field
from typing import Any, Generic, SupportsIndex, TypeVar

from .cache import Cacheable
from .fn import FnWithKwargs

K = TypeVar("K")
V = TypeVar("V")


@dataclass
class Entry(ABC):
    key: str

    @abstractmethod
    def __call__(self) -> Any:
        pass


class DictEntry(MutableMapping, Cacheable, Generic[K]):
    _data: dict[K, Entry]
    _resolve: bool

    def __init__(self, data: dict[K, Entry] | None = None, *, resolve: bool = True) -> None:
        super().__init__()

        self._data = data or {}
        self._resolve = resolve

    def __setitem__(self, key: K, value: Any) -> None:
        if not isinstance(value, Entry):
            msg = f"Value must be an Entry instance, got {type(value)}"
            raise TypeError(msg)

        self._data[key] = value
        self._update_id()

    def __getitem__(self, key: K) -> Any:
        value = self._data[key]

        if not isinstance(value, Entry):
            msg = f"Value must be an Entry instance, got {type(value)}"
            raise TypeError(msg)

        if self._resolve:
            return value.__call__()

        return value

    def __delitem__(self, key: K) -> None:
        self._data.__delitem__(key)
        self._update_id()

    def __iter__(self) -> Generator[K]:
        yield from self._data

    def __len__(self) -> int:
        return self._data.__len__()

    def __contains__(self, key: K) -> bool:
        return self._data.__contains__(key)


class ListEntry(MutableSequence, Cacheable):
    _data: list[Entry]
    _resolve: bool

    def __init__(self, data: list[Entry] | None = None, *, resolve: bool = True) -> None:
        super().__init__()

        self._data = data or []
        self._resolve = resolve

    def __setitem__(self, i: SupportsIndex, value: Any) -> None:
        if isinstance(value, Iterable):
            for value_i in value:
                if not isinstance(value_i, Entry):
                    msg = f"Value must be an Entry instance, got {type(value_i)}"
                    raise TypeError(msg)

        elif not isinstance(value, Entry):
            msg = f"Value must be an Entry instance, got {type(value)}"
            raise TypeError(msg)

        self._data[i] = value
        self._update_id()

    def __getitem__(self, i: SupportsIndex) -> Any:
        value = self._data.__getitem__(i)

        if isinstance(value, Iterable):
            new_values = []
            for value_i in value:
                if not isinstance(value_i, Entry):
                    msg = f"Value must be an Entry instance, got {type(value_i)}"
                    raise TypeError(msg)

                if self._resolve:
                    new_values.append(value_i.__call__())
                else:
                    new_values.append(value_i)

            return new_values

        if isinstance(value, Entry):
            if self._resolve:
                return value.__call__()

            return value

        msg = f"Value must be an Entry instance, got {type(value)}"
        raise TypeError(msg)

    def __delitem__(self, i: SupportsIndex) -> None:
        self._data.__delitem__(i)
        self._update_id()

    def __len__(self) -> int:
        return self._data.__len__()

    def insert(self, index: SupportsIndex, value: Any) -> None:
        if not isinstance(value, Entry):
            msg = f"Value must be an Entry instance, got {type(value)}"
            raise TypeError(msg)

        self._data.insert(index, value)
        self._update_id()


@dataclass
class FieldEntry(Entry, Generic[V]):
    value: V

    def __call__(self) -> V:
        return self.value


@dataclass
class ModuleEntry(Entry):
    obj: Any
    call: Any
    lazy: bool
    args: DictEntry[str] | ListEntry | None = None
    cache: bool = True
    fn: FnWithKwargs = field(init=False)
    bucket: dict[str] = field(init=False)

    def __post_init__(self) -> None:
        self.bucket = {}

        if self.call is False:
            return

        kwargs = {}
        args = ()

        if isinstance(self.args, DictEntry):
            kwargs = self.args
        elif isinstance(self.args, ListEntry):
            args = self.args

        if self.call is True:
            if not callable(self.obj):
                msg = f"'{self.obj}' is not callable"
                raise TypeError(msg)

            self.fn = FnWithKwargs(fn=self.obj, args=args, kwargs=kwargs)
            return

        if not hasattr(self.obj, self.call):
            msg = f"'{self.obj}' has no attribute '{self.call}'"
            raise AttributeError(msg)

        fn = getattr(self.obj, self.call)

        if not callable(fn):
            msg = f"'{fn}' is not callable"
            raise TypeError(msg)

        self.fn = FnWithKwargs(fn=fn, args=args, kwargs=kwargs)

    def __call__(self) -> Any | FnWithKwargs:
        if self.call is False:
            return self.obj

        if self.lazy:
            return self.fn

        if not self.cache:
            return self.fn.__call__()

        uid = None

        if self.args is not None:
            uid = self.args.uid

        if uid not in self.bucket:
            self.bucket[uid] = self.fn.__call__()

        return self.bucket[uid]



from pathlib import Path

from kaizo import ConfigParser

cache_py = """
num_call = 0
def fn():
    global num_call

    num_call+=1
    return num_call
"""

cache_config = """
local: main.py
with_cache:
  module: local
  source: fn
  cache: true
without_cache:
  module: local
  source: fn
  cache: false
"""


def test_cache(tmp_path: Path) -> None:
    module = tmp_path / "main.py"
    module.write_text(cache_py)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(cache_config)

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    with_cache_1 = out["with_cache"]
    with_cache_2 = out["with_cache"]
    without_cache_1 = out["without_cache"]
    without_cache_2 = out["without_cache"]

    assert with_cache_1 == with_cache_2
    assert without_cache_2 == without_cache_1 + 1



from pathlib import Path

import pytest

from kaizo import ConfigParser

dummy_config = """
dummy: this_is_dummy
"""

missing_local_config = """
local: missing.py
"""

module_not_exist_config = """
bad:
  module: does_not_exist
  source: thing
"""

invalid_import_type_config = """
import: not_a_dict
"""

missing_import_file_config = """
import:
  miss: missing_module.yml
"""

module_not_given_config = """
bad_ref: missing_module.{some_key}
"""

keyword_not_found_config = """
import:
  dummy: dummy.yml
bad_ref: missing_module.{some_key}
"""

entry_not_found_config = """
import:
  dummy: dummy.yml
bad_key: dummy.{not_here}
"""

local_entry_not_found_config = """
bad_key: .{not_here}
"""


def test_missing_local_module(tmp_path: Path) -> None:
    cfg = tmp_path / "cfg.yml"
    cfg.write_text(missing_local_config)

    with pytest.raises(FileNotFoundError):
        ConfigParser(cfg)


def test_import_error(tmp_path: Path) -> None:
    cfg = tmp_path / "cfg.yml"
    cfg.write_text(module_not_exist_config)

    parser = ConfigParser(cfg)

    with pytest.raises(ImportError):
        parser.parse()


def test_invalid_import_type(tmp_path: Path) -> None:
    cfg = tmp_path / "cfg.yml"
    cfg.write_text(invalid_import_type_config)

    with pytest.raises(TypeError):
        ConfigParser(cfg)


def test_missing_import_file(tmp_path: Path) -> None:
    cfg = tmp_path / "cfg.yml"
    cfg.write_text(missing_import_file_config)

    with pytest.raises(FileNotFoundError):
        ConfigParser(cfg)


def test_module_not_given(tmp_path: Path) -> None:
    cfg = tmp_path / "cfg.yml"
    cfg.write_text(module_not_given_config)

    parser = ConfigParser(cfg)

    with pytest.raises(ValueError, match="import module is not given"):
        parser.parse()


def test_keyword_not_found(tmp_path: Path) -> None:
    dummy_module = tmp_path / "dummy.yml"
    dummy_module.write_text(dummy_config)

    cfg = tmp_path / "cfg.yml"
    cfg.write_text(keyword_not_found_config)

    parser = ConfigParser(cfg)

    with pytest.raises(ValueError, match="keyword not found, got missing_module"):
        parser.parse()


def test_entry_not_found(tmp_path: Path) -> None:
    dummy_module = tmp_path / "dummy.yml"
    dummy_module.write_text(dummy_config)

    cfg = tmp_path / "cfg.yml"
    cfg.write_text(entry_not_found_config)

    parser = ConfigParser(cfg)

    with pytest.raises(KeyError, match="entry not found, got not_here"):
        parser.parse()


def test_local_entry_not_found(tmp_path: Path) -> None:
    cfg = tmp_path / "cfg.yml"
    cfg.write_text(local_entry_not_found_config)

    parser = ConfigParser(cfg)

    with pytest.raises(KeyError, match="entry not found, got not_here"):
        parser.parse()



from pathlib import Path

from kaizo import ConfigParser

X = 2
Y = 3
RESULT = X + Y

main_py = """
def add(x, y):
    return x + y
"""

relative_config = f"""
local: main.py
run:
  module: local
  source: add
  call: true
  args:
    x: {X}
    y: {Y}
"""

absolute_config = f"""
local: {{tmp_path}}/main.py
run:
  module: local
  source: add
  call: true
  args:
    x: {X}
    y: {Y}
"""


def test_relative_local_module(tmp_path: Path) -> None:
    module = tmp_path / "main.py"
    module.write_text(main_py)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(relative_config)

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    assert out["run"] == RESULT


def test_absolute_local_module(tmp_path: Path) -> None:
    module = tmp_path / "main.py"
    module.write_text(main_py)

    cfg_file = tmp_path / "cfg.yml"
    cfg_file.write_text(absolute_config.format(tmp_path=tmp_path))

    parser = ConfigParser(cfg_file)
    out = parser.parse()

    assert out["run"] == RESULT



import os
from pathlib import Path
from types import ModuleType
from typing import Any

import yaml

from .plugins import Plugin, PluginMetadata
from .utils import (
    DictEntry,
    Entry,
    FieldEntry,
    FnWithKwargs,
    ListEntry,
    ModuleEntry,
    ModuleLoader,
    Storage,
    extract_variable,
)


class ConfigParser:
    config: dict[str]
    local: ModuleType | None
    storage: dict[str, Storage]
    kwargs: dict[str]
    modules: dict[str, "ConfigParser"] | None
    plugins: dict[str, FnWithKwargs[Plugin]] | None

    def __init__(self, config_path: str | Path, kwargs: dict[str] | None = None) -> None:
        root, _ = os.path.split(config_path)

        root = Path(root)

        self.storage = {}
        self.kwargs = kwargs or {}

        with Path.open(config_path) as file:
            self.config = yaml.safe_load(file)

        if "local" in self.config:
            local_path = Path(self.config.pop("local"))

            if not local_path.is_absolute():
                local_path = root / local_path

            self.local = ModuleLoader.load_python_module(local_path)
        else:
            self.local = None

        if "import" in self.config:
            modules = self.config.pop("import")

            if not isinstance(modules, dict):
                msg = f"import module should be a dict, got {type(modules)}"
                raise TypeError(msg)

            self.modules = self._import_modules(root, modules, kwargs)
        else:
            self.modules = None

        if "plugins" in self.config:
            plugins = self.config.pop("plugins")

            if not isinstance(plugins, dict):
                msg = f"plugins should be a dict, got {type(plugins)}"
                raise TypeError(msg)

            self.plugins = self._import_plugins(plugins)
        else:
            self.plugins = None

    def _import_modules(
        self,
        root: Path,
        modules: dict[str, str],
        kwargs: dict[str] | None = None,
    ) -> dict[str, "ConfigParser"]:
        module_dict = {}

        for module_name, module_path_str in modules.items():
            module_path = Path(module_path_str)

            if not module_path.is_absolute():
                module_path = root / module_path

            parser = ConfigParser(module_path, kwargs)
            parser.parse()

            module_dict[module_name] = parser

        return module_dict

    def _import_plugins(
        self,
        plugins: dict[str],
    ) -> dict[str, FnWithKwargs[Plugin]]:
        metadata = PluginMetadata()
        plugin_dict = {}

        for plugin_name, plugin_module in plugins.items():
            plugin_path = f"kaizo.plugins.{plugin_name}"

            if isinstance(plugin_module, dict):
                source = plugin_module.get("source")
                args = plugin_module.get("args", {})

                resolved_args = self._resolve_args(plugin_name, args)
                metadata.args = resolved_args

                if source is None:
                    msg = f"source is required for {plugin_name} plugin"
                    raise ValueError(msg)

                plugin = ModuleLoader.load_object(plugin_path, source)

            elif isinstance(plugin_module, str):
                plugin = ModuleLoader.load_object(plugin_path, plugin_module)

            else:
                msg = f"plugin {plugin_name} is not a valid type"
                raise TypeError(msg)

            if not issubclass(plugin, Plugin):
                msg = f"loaded {plugin_name} is not a `Plugin`"
                raise TypeError(msg)

            obj = FnWithKwargs[Plugin](fn=plugin.dispatch, kwargs={"metadata": metadata})

            plugin_dict[plugin_name] = obj

        return plugin_dict

    def _load_symbol_from_module(self, module_path: str, symbol_name: str) -> Any:
        if module_path == "local":
            if self.local is None:
                msg = "local module is not given"
                raise ValueError(msg)

            return getattr(self.local, symbol_name)

        if module_path == "plugin":
            if self.plugins is None:
                msg = "plugins are not given"
                raise ValueError(msg)

            obj = self.plugins.get(symbol_name)

            if obj is None:
                msg = f"plugin {symbol_name} not found"
                raise ValueError(msg)

            return obj.__call__()

        return ModuleLoader.load_object(module_path, symbol_name)

    def _resolve_from_storage(
        self,
        storage: dict[str, Storage],
        *,
        key: str,
        entry_key: str | None,
        entry_sub_key: str,
    ) -> Entry | None:
        storage_key = entry_key
        storage_key_to_fetch = entry_sub_key or None

        if storage_key is None:
            storage_key = entry_sub_key
            storage_key_to_fetch = None

        if not storage_key:
            storage_key = key

        storage_i = storage.get(storage_key)

        if storage_i is None:
            return None

        return storage_i.get(storage_key_to_fetch)

    def _resolve_string(self, key: str, entry: str) -> Entry:
        entry_module, entry_key, entry_sub_key = extract_variable(entry)

        if entry_module is None:
            return FieldEntry(key=key, value=entry)

        if not entry_module:
            if entry_sub_key in self.kwargs:
                return FieldEntry(key=entry_sub_key, value=self.kwargs[entry_sub_key])

            parsed_entry = self._resolve_from_storage(
                self.storage,
                key=key,
                entry_key=entry_key,
                entry_sub_key=entry_sub_key,
            )

        else:
            if self.modules is None:
                msg = "import module is not given"
                raise ValueError(msg)

            module = self.modules.get(entry_module)

            if module is None:
                msg = f"keyword not found, got {entry_module}"
                raise ValueError(msg)

            parsed_entry = self._resolve_from_storage(
                module.storage,
                key=key,
                entry_key=entry_key,
                entry_sub_key=entry_sub_key,
            )

        if parsed_entry is None:
            msg = f"entry not found, got {entry_sub_key}"
            raise KeyError(msg)

        return parsed_entry

    def _resolve_list(self, key: str, entry: list) -> FieldEntry[list]:
        return FieldEntry(
            key=key,
            value=ListEntry([self._resolve_entry(key, e) for e in entry]),
        )

    def _resolve_args(self, key: str, args: Any) -> DictEntry[str] | ListEntry:
        if key not in self.storage:
            self.storage[key] = Storage.init()

        if isinstance(args, dict):
            resolved = DictEntry()
            for k, v in args.items():
                value = (
                    FieldEntry(key=key, value=self.kwargs[k])
                    if k in self.kwargs
                    else self._resolve_entry(key, v)
                )

                resolved[k] = value
                self.storage[key].set(k, value)

        if isinstance(args, list):
            resolved = ListEntry()
            for v in args:
                resolved.append(self._resolve_entry(key, v))

        return resolved

    def _resolve_dict(self, key: str, entry: dict[str]) -> Entry:
        module_path = entry.get("module")
        symbol_name = entry.get("source")

        if module_path is None or symbol_name is None:
            return FieldEntry(
                key=key,
                value=DictEntry({k: self._resolve_entry(key, v) for k, v in entry.items()}),
            )

        call = entry.get("call", True)
        lazy = entry.get("lazy", False)
        args = entry.get("args", {})
        cache = entry.get("cache", True)

        obj = self._load_symbol_from_module(module_path, symbol_name)

        resolved_args = self._resolve_args(key, args)

        return ModuleEntry(
            key=key,
            obj=obj,
            call=call,
            lazy=lazy,
            args=resolved_args,
            cache=cache,
        )

    def _resolve_entry(self, key: str, entry: Any) -> Entry:
        if isinstance(entry, str):
            return self._resolve_string(key, entry)

        if isinstance(entry, list):
            return self._resolve_list(key, entry)

        if isinstance(entry, dict):
            return self._resolve_dict(key, entry)

        return FieldEntry(key=key, value=entry)

    def parse(self) -> DictEntry[str]:
        res = DictEntry()

        for k in self.config:
            if k not in self.storage:
                self.storage[k] = Storage.init()

            value = self._resolve_entry(k, self.config[k])

            res[k] = value
            self.storage[k].value = value

        return res



import importlib
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from types import ModuleType
from typing import Any


class ModuleLoader:
    @staticmethod
    def load_python_module(path: Path) -> ModuleType:
        if not path.is_file():
            msg = f"Local Python file not found: {path}"
            raise FileNotFoundError(msg)

        module_name = path.stem
        spec = spec_from_file_location(module_name, path)
        if spec is None or spec.loader is None:
            msg = f"Failed to load module from: {path}"
            raise ImportError(msg)

        module = module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    @staticmethod
    def load_object(module_path: str, object_name: str) -> Any:
        try:
            module = importlib.import_module(module_path)
            if not hasattr(module, object_name):
                msg = f"Module '{module_path}' has no attribute '{object_name}'"
                raise AttributeError(msg)
            return getattr(module, object_name)
        except ModuleNotFoundError as e:
            msg = f"Could not import module '{module_path}': {e}"
            raise ImportError(msg) from e



import uuid


class Cacheable:
    _id: str

    def __init__(self) -> None:
        self._id = uuid.uuid4().hex

    def _update_id(self) -> None:
        self._id = uuid.uuid4().hex

    @property
    def uid(self) -> str:
        return self._id



from dataclasses import dataclass
from typing import Self

from .entry import DictEntry, Entry


@dataclass
class Storage:
    value: Entry | None
    items: DictEntry[str]

    @staticmethod
    def init() -> Self:
        return Storage(
            value=None,
            items=DictEntry(resolve=False),
        )

    def get(self, key: str | None = None) -> Entry | None:
        if key is None:
            return self.value

        return self.items.get(key)

    def set(self, key: str, value: Entry) -> None:
        self.items[key] = value
